from typing import List

import xlsxwriter
import pandas as pd
import numpy as np

from .commons import run_requests, structure_sheets, structure_sheet
from ..commons.consts import RED_TO_GREEN_GRADIENT
from ..services.services import Services


def convert_to_xlsx(
    input_path: str,
    services: Services,
    workers: int = 1,
    output_path: str = None,
    normalize: bool = None,
):
    """Sends requests to defined Hypatos services for every invoice found in given
    directory. Merges results off all services into one Excel file and writes to
    project directory. Every key, value pair from Invoice Extractor will flattened
    into the following tuple:

        key = (value, probability)

    Arguments:
        path {str} -- Path to the directory containing invoices, uses rglob.
        extractor_endpoint {str} -- Url of Hypatos extractor service.

    Keyword Arguments:
        token {str} -- API token. (default: {None})
        workers {int} -- Amount of multithread. (default: {1})
    """
    # Get requests result
    response = run_requests(workers, input_path, services)

    # Structure response into sheets (dict) containing records [{...}, {...}]
    sheets = structure_sheets(response)

    # Normalize probability
    if normalize and (services.headers.get("probabilities") == "true"):
        for sheet in sheets.copy():
            sheets[sheet] = normalize_probability(sheets[sheet])

    # Write sheets into workbook
    write_workbook(output_path, sheets)


def write_workbook(output_path: str, sheets: List[dict]):
    # Init workbook/sheets
    workbook = xlsxwriter.Workbook(output_path)

    # Styling
    header_format = workbook.add_format({"bold": True})
    number_format = workbook.add_format({"num_format": "0.00"})
    probability_formats = [
        workbook.add_format({"bold": True, "bg_color": color})
        for color in RED_TO_GREEN_GRADIENT
    ]

    for sheet in sheets:
        workbook.add_worksheet(sheet)
        write_sheet(
            workbook.get_worksheet_by_name(sheet),
            structure_sheet(sheets[sheet]),
            header_format,
            number_format,
            probability_formats,
        )

    workbook.close()


def write_sheet(
    worksheet, records: list, header_format, number_format, probability_formats: list
):
    """Write items to workbook sheet.

    Arguments:
        worksheet {[type]} -- [description]
        records {[type]} -- [description]
        bold_header {[type]} -- [description]
        red_to_green_formats {[type]} -- [description]
    """
    for idx, row in enumerate(records):
        count = 0
        for key, value in row.copy().items():
            worksheet.write(0, count, key, header_format)
            if not isinstance(value, tuple):
                records[idx][key] = (value, None, None)

            column_value, probability, _ = records[idx][key]
            value_format = number_format if isinstance(column_value, float) else None

            if pd.notna(probability):
                color_idx = int((len(probability_formats) - 1) * probability)
                color = probability_formats[color_idx]
                worksheet.write(idx + 1, count, column_value, color)
            else:
                worksheet.write(idx + 1, count, column_value, value_format)

            count += 1

    for i, width in enumerate(get_col_widths(pd.DataFrame.from_records(records))):
        worksheet.set_column(i, i, width)


def get_col_widths(invoices, max_col_width=70):
    """ Get max length for every column and below max_column_width """
    return [
        min(
            max(
                [len(str(s[0])) for s in invoices[col].values if s is not np.nan]
                + [len(col)]
            ),
            max_col_width,
        )
        for col in invoices.columns
    ]


def normalize_probability(records: list):
    """ Normalize the probabilities for every column inbetween range(min, max) of
    particulair column.

    Args:
        invoices ([type]): [description]

    Returns:
        [type]: [description]
    """
    # Read
    probabilities = {}
    for idx, invoice in enumerate(records):
        probabilities[idx] = {k: v[1] for k, v in invoice.items()}

    # Normalize
    df = pd.DataFrame.from_dict(probabilities, orient="index")
    for col in df.columns:
        if not df[col].isnull().all():
            min_value, max_value = (
                df[col].min(),
                df[col].max(),
            )
            if min_value == max_value:
                df[col] = pd.Series([np.nan] * len(df[col]))
            else:
                df[col] = (df[col] - min_value) / (max_value - min_value)

    # Write
    normalized = df.to_dict("records")
    for idx, invoice in enumerate(records.copy()):
        for col in invoice:
            if invoice[col]:
                triplet = list(records[idx][col])
                triplet[1] = normalized[idx][col]
                records[idx][col] = tuple(triplet)

    return records



import os
import json

from concurrent.futures import ThreadPoolExecutor, as_completed

import click

from ..services.requests import extract_invoice
from .commons import read_pdf, get_filenames


def convert_to_json(
    input_path, services, workers, output_path: str = None, skip: bool = False
):
    files = get_filenames(input_path, output_path, skip)
    skipped_files = []

    with ThreadPoolExecutor(max_workers=workers) as exe:
        jobs = {
            exe.submit(
                extract_invoice,
                read_pdf(file_path),
                services.extractor_endpoint,
                file_extension,
                services.get_token,
                services.headers,
            ): file_path
            for file_path, file_extension in files.items()
        }
        label = f"Converting {len(jobs)} invoices"
        with click.progressbar(jobs, label=label) as bar:
            for idx, future in enumerate(as_completed(jobs)):
                file_name = jobs[future].split("/")[-1]
                try:
                    response = future.result(timeout=300)

                except Exception as e:
                    skipped_files.append((file_name, e))
                    continue

                file_path = f"{file_name.split('.')[0]}.json"

                if output_path:
                    file_path = os.path.join(output_path, file_path)
                    os.makedirs(os.path.dirname(file_path), exist_ok=True)

                with open(file_path, "w") as f:
                    json.dump(response, f)

                bar.update(1)

    return skipped_files



from typing import Callable, Dict

import requests

from ..commons.error_handling import handle_status_code


@handle_status_code
def request_token(token_endpoint, username, password):
    """ Request invoice extractor token """
    response = requests.post(
        url=token_endpoint,
        data={
            "username": username,
            "password": password,
            "grant_type": "password",
            "client_id": "extractionapi",
        },
        headers={"Content-Type": "application/x-www-form-urlencoded"},
    )
    response.raise_for_status()
    return response.json()["access_token"]


@handle_status_code
def extract_invoice(
    file,
    extractor_endpoint,
    content_type,
    get_token: Callable[[], str] = None,
    params: Dict = {},
    timeout: int = 1800,
):
    """ Request invoice extractor """
    params.setdefault("ingnoreNull", "false")
    response = requests.post(
        url=extractor_endpoint,
        data=file,
        params=params,
        timeout=timeout,
        headers={
            "Content-type": content_type,
            "Authorization": f"Bearer {get_token()}",
        },
    )

    response.raise_for_status()
    return response.json()



from .convert.xlsx import convert_to_xlsx, write_workbook
from .convert.csv import convert_to_csv
from .convert.json import convert_to_json
from .services.services import Services
from .commands.compare import ModelComparer


__version__ = "0.3.28"
__all__ = [
    "convert_to_csv",
    "convert_to_xlsx",
    "convert_to_json",
    "Services",
    "write_workbook",
    "ModelComparer",
]



import click
from halo import Halo

from .. import convert_to_csv, Services


@click.command(context_settings=dict(max_content_width=200))
@click.argument("path", type=click.Path(exists=True))
@click.option("-w", "--workers", default=1, show_default=True, help="amount of workers")
@click.option(
    "-p",
    "--probability",
    default=False,
    show_default=True,
    help="include probability",
    is_flag=True,
)
@click.pass_context
def to_csv(ctx, path, workers, probability):
    """ Convert invoice to csv """
    # Services available
    spinner = Halo(spinner="dots")
    spinner.start()
    services = Services(ctx, check_up=True)
    spinner.succeed(f"Found dir: {path}\n ")

    # Conversion
    convert_to_csv(path, services, workers, path)
    spinner.succeed("Converted invoice(s) to csv")



import os
import csv
from copy import deepcopy

from .commons import run_requests, structure_sheets, structure_sheet
from ..services.services import Services


def convert_to_csv(
    input_path: str,
    services: Services,
    workers: int = 1,
    output_path: str = None,
    probability=False,
):
    # Get requests result
    result = run_requests(workers, input_path, services)
    processed_dir_name = os.path.normpath(output_path).split(os.path.sep)[-1]

    # Structure result
    sheets = structure_sheets(result)

    for sheet in sheets:
        write_csv(f"{processed_dir_name}_{sheet}.csv", structure_sheet(sheets[sheet]))


def write_csv(file_name, records):
    set_value(records)
    with open(file_name, mode="w") as f:
        writer = csv.DictWriter(
            f, fieldnames=records[0].keys(), extrasaction="ignore", delimiter=";",
        )
        writer.writeheader()
        for row in records:
            writer.writerow(row)


def set_value(records):
    for idx, row_items in enumerate(deepcopy(records)):
        for field, value in row_items.items():
            records[idx][field] = value[0]



import pandas as pd
from functools import partial
from difflib import SequenceMatcher
import hashlib
import os

import logging
import click


class ModelComparer:

    def __init__(self, file_path_1: str, file_path_2: str):
        self.file_path_1 = file_path_1
        self.file_path_2 = file_path_2

    def load_files(self):

        def _load_file(file_path: str) -> dict:
            def _check_extesion() -> str:
                file_name = os.path.basename(file_path)
                extension = os.path.splitext(file_name)[1]
                return extension

            if _check_extesion()[:3] == ".xl":
                return dict(pd.read_excel(file_path, None))

            else:
                raise TypeError(
                    f"{file_path}: Type {_check_extesion()} is currently not supported.\nSupported types: xlsx, xls, "
                    f"xlsm")

        self.model_1 = _load_file(self.file_path_1)
        self.model_2 = _load_file(self.file_path_2)

    def check_files(self):

        def _check_dataframe(df, ws_name, path):
            '''Checks if dataframe has a column named file_name which is required to perform other calculations'''

            assert "file_name" in df.columns, f"Worksheet '{ws_name}' in file '{path}' has no column 'file_name' which is required to perform further operations. \n Please check the names of columns and run the program once again."

        [_check_dataframe(df, name, self.file_path_1) for name, df in self.model_1.items()]
        [_check_dataframe(df, name, self.file_path_2) for name, df in self.model_2.items()]

    @staticmethod
    def export_comparison_to_excel(coparisons: dict, file_path: str):
        def _format_excel(df: pd.DataFrame) -> pd.DataFrame:

            def _highlight_records(val):
                val = -1 if val == '' else val

                if val > 0:
                    color = '#FFB6C1'
                    return f'background-color: {color}'

            diff_cols = [col for col in df.columns if 'diff_' in col]

            df = df.reset_index(drop=True).style.set_properties(**{
                'background-color': '#FFFFFF',
            }).applymap(_highlight_records, subset=list(diff_cols))

            return df

        with pd.ExcelWriter(file_path, engine='openpyxl', mode='wa') as writer:
            for k, v in coparisons.items():
                v = _format_excel(v)
                v.to_excel(writer, sheet_name=k)
            writer.save()

    # TODO: this function should be simplified
    def compare_differences(self) -> dict:  # noqa

        def _apply_difference(s: pd.DataFrame, a: str, b: str, tolerance=0.05) -> float:
            '''
            Applies difference ratio (from 0 to 1) between each record on given pandas DataFrame using Levenshtein distance.
            Difference is calculated between x column and y column.
            '''

            try:
                if any([type(s[a]) is str, type(s[b]) is str]):
                    # For comparing strings
                    similarity = SequenceMatcher(None, str(s[a]).lower(), str(s[b]).lower()).ratio()

                    if similarity < 0.6:
                        difference = 1
                    else:
                        difference = round(1.00 - similarity, 2)

                else:
                    # For comparing values
                    higher_value = float(max([s[a], s[b]]))
                    lower_value = float(min([s[a], s[b]]))
                    if lower_value == 0:
                        lower_value = 0.01

                    difference = round(abs(1 - higher_value / lower_value), 2)

                    if difference > tolerance:
                        difference = 1

            except KeyError:
                difference = 0 if s[a] == '' else 1

            return difference

        def _create_difference_for_each_column(df: pd.DataFrame) -> pd.DataFrame:
            for _ in df.columns:

                if _[-2:] == "_x":  # for each _x column (ground truth)
                    x_column = _  # column which has _x suffix (from ground truth)
                    new_column = "diff_" + _[:-2]  # column which has diff_ prefix with calculated difference
                    y_column = _[:-2] + "_y"  # column which has _y suffix (some dataset)

                    # Calculate Levenshtein difference between each record in column x and column y
                    df[new_column] = df.apply(partial(_apply_difference, a=x_column, b=y_column), axis=1)

            return df

        def _diff_summary_by_row(df: pd.DataFrame) -> pd.DataFrame:
            """Provide difference summary for each row"""
            arr = []
            for column in df.columns:
                if column[:5] == "diff_":
                    arr.append(column)
            return df[arr].sum(axis=1)

        def _sort_columns(df: pd.DataFrame) -> list:
            """Use this function to sort columns with order x, y, difference"""
            arr = ["file_name", "distance", "diff_total"]
            for _ in df.columns:

                if _[-2:] == "_x":
                    arr.append(_)
                    arr.append(_[:-2] + "_y")
                    arr.append("diff_" + _[:-2])

            return arr

        def _summary(df_1: pd.DataFrame, df_2: pd.DataFrame) -> pd.DataFrame:
            """
            Provide summary based on original df and results.
            df_1: ground_truth
            df_2: extraction
            """
            number_of_documents = df_1.shape[0]
            item_categories = len(
                [column for column in df_2.columns if column.startswith("diff_") and column != "diff_total"])
            data_points = number_of_documents * item_categories
            percent_diff = round(self.total_difference / data_points, 2)

            data = {
                1: ["Number of documents", None, number_of_documents],
                2: ["Number of compared columns", None, item_categories],
                3: ["Compared data points", None, data_points],
                4: ["Overall difference in percent", None, percent_diff],
                5: ["Overall accuracy in percent", None, 1 - percent_diff]
            }

            summary = pd.DataFrame.from_dict(data, orient='index', columns=['file_name', 'distance', 'diff_total'])

            df_2 = df_2.append(summary)

            return df_2

        def _compare_single(df_1, df_2):
            simple_left_join = pd.merge(df_1, df_2, how='left', left_on="file_name", right_on="file_name")
            simple_left_join = _create_difference_for_each_column(simple_left_join)
            return simple_left_join

        def _compare_multi(dataframe_name: str) -> pd.DataFrame:

            def _apply_best_matches(df_1: pd.DataFrame, df_2: pd.DataFrame, file_name: str,
                                    left_matching_index="file_name", right_matching_index="file_name") -> pd.DataFrame:
                '''
                Based on two dataframes, perform left-join type matching using Levenshtein lookup for each row.
                '''

                filtered_file_name = df_1[df_1.file_name == file_name].copy()

                # prepare indexes for artificial_keys
                filtered_file_name['index_x'] = [hashlib.md5(val.encode('utf-8')).hexdigest() for val in
                                                 filtered_file_name['artificial_key']]
                df_2['index_y'] = [hashlib.md5(val.encode('utf-8')).hexdigest() for val in df_2['artificial_key']]

                # create cartesian product for file
                cartesian = pd.merge(filtered_file_name, df_2, left_on=left_matching_index,
                                     right_on=right_matching_index)

                if not cartesian.empty:

                    # how many records we have on _x
                    records_on_x = filtered_file_name["file_name"].count()
                    # how many records we have on _y
                    records_on_y = df_2[df_2.file_name == file_name]["file_name"].count()

                    cartesian["distance"] = cartesian.apply(
                        partial(_apply_difference, a="artificial_key_x", b="artificial_key_y"), axis=1)

                    def _sudoku_matcher(df: pd.DataFrame, per_index: str) -> pd.DataFrame:
                        '''

                        Creates a dataframe based on a list of matches that are possible (e.g. matching A1 to A5 and
                        A2 to A4 is possible, while matching A1 to A5 and A1 to A4 is not possible due to A1 being
                        matched twice)

                        '''

                        # introduce empty DataFrame at the beginning
                        best_results = pd.DataFrame()

                        # copy a dataset that we will working with
                        dataset = df

                        # execute in loop until dataset would have no possible matches
                        while not dataset.empty:
                            # collect the best match
                            best_match = dataset.loc[dataset['distance'].idxmin()]
                            best_results = best_results.append(best_match)

                            index_to_be_removed = best_match.index_x
                            corresponding_field = best_match.index_y

                            # perform elimination within dataset
                            dataset = dataset.drop(dataset[dataset["index_x"] == index_to_be_removed].index)
                            dataset = dataset.drop(dataset[dataset["index_y"] == corresponding_field].index)

                        # TODO: this should be moved to unit tests
                        # assertions checks that there should be no duplicates after the executing the loop
                        # within matched dataset

                        assert not dataset['index_y'].duplicated().any(), dataset
                        assert not dataset['index_x'].duplicated().any(), dataset
                        assert not best_results['index_x'].duplicated().any(), (file_name, best_results.index_x)
                        assert not best_results['index_y'].duplicated().any(), (file_name, best_results.index_y)

                        # if there are more items on a left dataset than in the right dataset
                        # collect not matched records from left dataset (x) and place them in a results as non matched
                        if per_index == "index_x":
                            not_matched_results = df[~df.index_x.isin(best_results.index_x)]
                            if not not_matched_results.empty:
                                dummy_indexes = not_matched_results['index_x'].unique()
                                dummy_records = filtered_file_name[filtered_file_name.index_x.isin(dummy_indexes)]
                                old_columns = dummy_records.columns
                                new_columns = [
                                    column + "_x" if (column != "file_name" and column != "index_x") else column
                                    for column in old_columns]
                                dummy_records.columns = new_columns
                                best_results = best_results.append(dummy_records)

                        return best_results

                    matches = _sudoku_matcher(cartesian,
                                              'index_x') if records_on_x >= records_on_y else _sudoku_matcher(
                        cartesian, 'index_y')
                    matches.drop(["artificial_key_x"], axis=1, inplace=True)
                    matches.drop(["artificial_key_y"], axis=1, inplace=True)
                    matches.drop(["index_x"], axis=1, inplace=True)
                    matches.drop(["index_y"], axis=1, inplace=True)

                    return matches

                else:
                    matches = pd.merge(filtered_file_name, df_2, on='file_name', how='left')
                    matches.drop(["artificial_key_x"], axis=1, inplace=True)
                    return matches

            # Prepare data
            df_1 = self.model_1[dataframe_name]
            df_2 = self.model_2[dataframe_name]

            df_1.fillna("", inplace=True)
            df_2.fillna("", inplace=True)

            df_1['artificial_key'] = df_1.astype(str).values.sum(axis=1)
            df_2['artificial_key'] = df_2.astype(str).values.sum(axis=1)

            grouped_1 = df_1.groupby('file_name')['file_name'].count()

            # Apply best matches for each file
            results = pd.DataFrame()
            for k, v in grouped_1.to_dict().items():
                r = _apply_best_matches(df_1, df_2, k)
                results = results.append(r)

            results = _create_difference_for_each_column(results)
            return results

        def _line_items_statistics(dataframes: tuple) -> pd.DataFrame:

            df_1 = dataframes[0]
            df_2 = dataframes[1]

            group_1 = df_1.groupby(['file_name'])['row_number'].count()
            group_2 = df_2.groupby(['file_name'])['row_number'].count()

            df = pd.merge(group_1, group_2, how='outer', left_on='file_name', right_on='file_name')
            df.fillna(0, inplace=True)
            df['diff_count_of_items'] = df['row_number_x'] - df['row_number_y']
            df.rename(columns={'row_number_x': 'count_of_x_items', 'row_number_y': 'count_of_y_items'}, inplace=True)
            df['count_of_x_items'].astype('int')
            df['count_of_y_items'].astype('int')

            return df

        def run_excel_flow() -> dict:
            comparisons = {}

            def _match_worksheets() -> list:
                """
                Matches counterparts for each worksheets looking for exactly the same worksheet name in different workbook.
                [('invoice', 'invoice'), ('items', 'items'), ('payment_terms', 'payment_terms'), ('taxes', 'taxes')]

                The function will throw an warning if one of the worksheets would have different name for example:
                [('invoice', 'invoices')]

                The function will raise an error if there would be no counterparts in different file.
                found_ws == []
                """

                worksheets_in_model_1 = list(self.model_1.keys())
                worksheets_in_model_2 = list(self.model_2.keys())

                found_ws = []
                not_found_ws = []

                for ws in worksheets_in_model_1:
                    if ws in worksheets_in_model_2:
                        found_ws.append(ws)
                    else:
                        not_found_ws.append(ws)

                if not_found_ws:
                    logging.warning(
                        f"Skipping comparing {not_found_ws} worksheets. The program was not able to find their counterparts in the second workbook.")

                if found_ws:
                    logging.info(
                        f"Comparing {found_ws} worksheets...")
                    return found_ws

                assert found_ws, f"The program was not able to find any of {worksheets_in_model_1} worksheets in the second workbook."

            matched_worksheets = _match_worksheets()

            for worksheet in matched_worksheets:

                # If any dataframe starts with invoice
                if any([df.startswith("invoice") for df in worksheet]):
                    # Compare Single
                    comparison_results = _compare_single(self.model_1[worksheet[0]], self.model_2[worksheet[0]])
                else:
                    # Compare Multi
                    comparison_results = _compare_multi(worksheet)

                # Prepare total for each row
                comparison_results['diff_total'] = _diff_summary_by_row(comparison_results)

                # Sort columns
                comparison_results = pd.DataFrame(comparison_results, columns=_sort_columns(comparison_results))

                diff_cols = [col for col in comparison_results.columns if ('diff_' in col)]

                # Calculate total difference in percent
                number_of_records = self.model_1[worksheet].shape[0]

                # Calculate total difference
                sum_of_differences = round(comparison_results[diff_cols].sum(numeric_only=True, axis=0))
                percent_differences = round(sum_of_differences / number_of_records, 2)

                # Place total difference
                comparison_results.loc["diff_total"] = round(sum_of_differences, 2)

                # Place below percent difference
                comparison_results.loc["diff_percent"] = percent_differences

                comparison_results.at["diff_total", "file_name"] = "Total difference"
                comparison_results.at["diff_percent", "file_name"] = "Total difference in percent"

                self.total_difference = comparison_results.loc["diff_total", "diff_total"]
                comparison_name = worksheet
                comparison_results = _summary(self.model_1[comparison_name], comparison_results)

                comparisons[comparison_name] = comparison_results

            try:
                comparisons["line_items_statistics"] = _line_items_statistics(
                    (self.model_1["items"], self.model_2["items"]))
            except Exception:
                logging.warning(
                    "Skipping preparing line item statistics. Please check if worksheet 'items' exists in both workbooks")

            return comparisons

        return run_excel_flow()


@click.command(context_settings=dict(max_content_width=200))
@click.argument("file_path_1", type=click.Path(exists=True))
@click.argument("file_path_2", type=click.Path(exists=True))
@click.option(
    "-o",
    "--output",
    help="output path for xlsx file relative from current location (ends with .xlsx)",
    default='./comparison.xlsx',
)
@click.pass_context
def compare(ctx, file_path_1, file_path_2, output):
    """ compares extraction data between two excel files """

    mc = ModelComparer(file_path_1, file_path_2)
    mc.load_files()
    logging.info(f'IMPORTED EXCEL: {file_path_1} and {file_path_2}')
    comparisons = mc.compare_differences()
    mc.export_comparison_to_excel(comparisons, output)
    logging.info(f'EXPORTED EXCEL TO: {output}')



RED_TO_GREEN_GRADIENT = [
    "#FF0000",
    "#FF1100",
    "#FF2300",
    "#FF3400",
    "#FF4600",
    "#FF5700",
    "#FF6900",
    "#FF7B00",
    "#FF8C00",
    "#FF9E00",
    "#FFAF00",
    "#FFC100",
    "#FFD300",
    "#FFE400",
    "#FFF600",
    "#F7FF00",
    "#E5FF00",
    "#D4FF00",
    "#C2FF00",
    "#B0FF00",
    "#9FFF00",
    "#8DFF00",
    "#7CFF00",
    "#6AFF00",
    "#58FF00",
    "#47FF00",
    "#35FF00",
    "#24FF00",
    "#12FF00",
    "#00FF00",
]

default_endpoints = {
    "production": {
        "extractor": "https://api.hypatos.ai/v2/subscription/invoices",
        "authentication": "https://customers.hypatos.ai/auth/realms/hypatos/protocol/openid-connect/token",
    },
    "staging": {
        "extractor": "https://api.stage.hypatos.ai/v2/extract",
        "authentication": "https://customers.stage.hypatos.ai/auth/realms/hypatos/protocol/openid-connect/token",
    },
    "localhost": {"extractor": "http://localhost:8000/v2/extract"},
}



import os
import re
import json
import itertools

from concurrent.futures import ThreadPoolExecutor, as_completed
from pathlib import Path

import click
from filetype import guess_mime

from ..services.requests import extract_invoice


def run_requests(workers: int, path: str, services) -> dict:
    files = get_filenames(path)
    result = {}

    with ThreadPoolExecutor(max_workers=workers) as exe:
        jobs = {
            exe.submit(
                extract_invoice,
                read_pdf(file_path),
                services.extractor_endpoint,
                file_extension,
                services.get_token,
                services.headers,
            ): file_path
            for file_path, file_extension in files.items()
            if not file_extension.endswith("json")
        }
        label = f"Converting {len(jobs)} invoices"

        with click.progressbar(jobs, label=label) as bar:
            for idx, future in enumerate(as_completed(jobs)):
                file_name = jobs[future].split("/")[-1]

                try:
                    extracted_invoice = future.result(timeout=300)
                    result[idx] = flatten_invoice(extracted_invoice)
                except Exception as e:
                    result[idx] = {"error_message": (str(e), None, None)}
                finally:
                    result[idx]["file_name"] = (file_name, None, None)

                bar.update(1)

        json_jobs = {
            file_path: json.loads(file_path.read_bytes())
            for file_path, file_extension in files.items()
            if file_extension.endswith("json")
        }
        label = f"Merging {len(json_jobs)} invoices"

        with click.progressbar(json_jobs, label=label) as bar:
            for idx, (file_path, document) in enumerate(
                json_jobs.items(), start=len(jobs)
            ):
                result[idx] = flatten_invoice(document)
                result[idx]["file_name"] = (file_path.name, None, None)

                bar.update(1)

    if not result:
        quit("No files found in path")
    return result


def flatten_invoice(invoice):
    return_dict = dict()
    entities = invoice["entities"]
    probabilities = invoice.get("probabilities")

    def traverse_items(entities, probabilities, _dict, *prefix):
        for k, v in entities.items():
            if isinstance(v, dict):
                traverse_items(
                    entities[k],
                    probabilities[k] if probabilities else None,
                    return_dict,
                    k,
                )
            elif isinstance(v, list):
                # items, taxes, terms
                if all(isinstance(list_item, dict) for list_item in v):
                    for counter, list_item in enumerate(v):
                        temp_dict = {}
                        for item, value in list_item.items():
                            temp_dict[f"{k}_{item}_{counter}"] = value
                        traverse_items(
                            temp_dict,
                            probabilities[k][counter] if probabilities else None,
                            return_dict,
                        )
                # ibanAll enc
                elif all(isinstance(list_item, str) for list_item in v):
                    if v:
                        _dict[k] = []
                        for _, list_item in enumerate(v):
                            _dict[k].append(list_item)
                        _dict[k] = (", ".join(_dict[k]), None, None)
                    else:
                        _dict[k] = ("", None, None)
                # Undefined datastructure
                else:
                    pass
            else:
                try:
                    # dirty solution, assumes no invoice extractor response field got underscore
                    original_k = k.split("_")[-2]
                except IndexError:
                    original_k = k

                if prefix:
                    field_name = f"{prefix[0]}_{k}"
                else:
                    field_name = k

                if probabilities and original_k in probabilities:
                    if probabilities[original_k]:
                        _dict[field_name] = (
                            v if v else "",
                            probabilities[original_k],
                            None,
                        )
                    else:
                        _dict[field_name] = (v if v else "", None, None)
                else:
                    _dict[field_name] = (v if v else "", None, None)

    traverse_items(entities, probabilities, return_dict)
    return return_dict


def read_pdf(pdf_path: str) -> bytes:
    with open(pdf_path, "rb") as pdf:
        pdf = pdf.read()

    return pdf


def get_filenames(
    input_path: str, output_path: str = None, skipped: bool = False
) -> dict:
    """Find all filepaths needed to be processed

    Args:
        input_path (str): [description]
        output_path (str): [description]
        skipped (bool): [description]

    Returns:
        list: files found in input_path
    """
    abs_input_path = os.path.join(os.getcwd(), input_path)
    f_types = ("*.pdf", "*.tif", "*.tiff", "*.png", "*.jpg")

    found_files = []
    skippable_files = []

    for f_type in f_types:
        found_files.extend(Path(abs_input_path).rglob(f_type))
        found_files.extend(Path(abs_input_path).rglob(f_type.upper()))

    if skipped:
        skippable_files.extend(Path(output_path).rglob("*.json"))
        skippable_files = [f.stem for f in skippable_files]
        found_files = [x for x in found_files if x.stem not in skippable_files]

    # TODO: write fix for non mime recognized files
    # Currently only mime recognized files are processed
    found_files = [str(f) for f in found_files]

    if any([f.endswith(".json") for f in found_files]):
        return {f: guess_mime(f) for f in found_files if guess_mime(f)}
    else:
        return {
            **{f: guess_mime(f) for f in found_files if guess_mime(f)},
            **{f: "application/json" for f in Path(abs_input_path).rglob("*.json")},
        }


def structure_sheets(response: dict, header_sheetname="header"):
    """Structure the extracted invoices in different sheets.

    Arguments:
        response {[type]} -- [description]

    Returns:
        [type] -- [description]
    """
    single_cardinality = {header_sheetname: []}
    multi_cardinality = {}

    for idx, document in response.items():
        single_cardinality[header_sheetname].append(
            {
                "file_name": document["file_name"],
                "error_message": document.get("error_message", (None, None, None)),
            }
        )
        multi_items = {}

        for col, value in document.items():
            if col[-1].isdigit():
                item_num = int(col.split("_")[-1])
                label = col.split("_")[1]
                category = col.split("_")[0]
                multi_items.setdefault(category, [])

                if len(multi_items[category]) == item_num:
                    multi_items[category].append(
                        {
                            "file_name": document["file_name"],
                            "row_number": (item_num + 1, None, None),
                        }
                    )

                multi_items[category][item_num][label] = document[col]
            else:
                single_cardinality[header_sheetname][idx][col] = document[col]

        for category, rows in multi_items.items():
            for row in rows:
                multi_cardinality.setdefault(category, []).append(row)

    return {**single_cardinality, **multi_cardinality}


def get_output_path(input_path, output_path, extension="xlsx"):
    if not output_path:
        if input_path == ".":
            output_path = f"hycli.{extension}"
        else:
            output_path = (
                f"{os.path.normpath(input_path).split(os.path.sep)[-1]}.{extension}"
            )
    else:
        output_path = f"{os.path.join(os.getcwd(), output_path)}"
        if not os.path.isdir(os.path.sep.join(output_path.split(os.path.sep)[:-1])):
            quit("Output directory does not exist")

    return output_path


def structure_sheet(records):
    columns = set()

    # Collect all column names
    for record in records:
        columns = columns.union(set(record.keys()))

    for idx, record in enumerate(records.copy()):
        # Insert all missing columns into record
        record.update(
            {k: (None, None, None) for k in columns if k not in record.keys()}
        )

        # All snake_case keys
        record = {
            re.sub(r"(?<!^)(?=[A-Z])", "_", k).lower(): v for k, v in record.items()
        }

        # Reorder column order
        if "error_message" in record or "row_number" in record:
            records[idx] = {
                **dict(itertools.islice(record.items(), 2)),
                **dict(sorted(itertools.islice(record.items(), 2, None))),
            }
        else:
            records[idx] = {
                **dict(itertools.islice(record.items(), 1)),
                **dict(sorted(itertools.islice(record.items(), 1, None))),
            }

    return records



import click
from halo import Halo

from .. import convert_to_json, Services


@click.command(context_settings=dict(max_content_width=200))
@click.argument("path", type=click.Path(exists=True))
@click.option("-w", "--workers", default=3, show_default=True, help="amount of workers")
@click.option(
    "-o",
    "--output",
    help="output path for json files absolute or relative from current location.",
)
@click.option(
    "-s",
    "--skip",
    is_flag=True,
    default=False,
    help="Skip converting documents which are found to have been converted in output path.",
)
@click.pass_context
def to_json(ctx, path, workers, output, skip):
    """ Convert invoices to jsons """
    # Services available
    spinner = Halo(spinner="dots")
    spinner.start()
    services = Services(ctx, check_up=True)
    spinner.succeed(f"Found dir: {path}\n ")

    # Conversion
    skipped_files = convert_to_json(path, services, workers, output, skip)

    if skipped_files:
        spinner.fail("Could not convert all invoice(s) to json output.")
        for skipped_file in skipped_files:
            click.echo(f"{skipped_file[0]} - Error: {skipped_file[1]}")
    else:
        spinner.succeed("Converted all invoice(s) to json output.")



def handle_status_code(func):
    """ Decorator to handle http errors from requests lib.

    Arguments:
        func {[type]} -- Function which include requests func and response.raise_for_status()

        https://github.com/psf/requests/blob/2758124a13cddff7244b97b5ffe3fddabb90bc18/requests/status_codes.py

    Raises:
        Exception: Exception being raised in response.raise_for_status() will be raised in hycli.

    Returns:
        [type] -- [description]
    """

    def decorator(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as err:
            if err.response is not None:
                if err.response.status_code == 401:
                    raise Exception("API credentials not correct.")
                elif hasattr(err.response, "_content"):
                    raise Exception(str(err.response._content, "utf-8"))
                else:
                    raise Exception(err)
            else:
                raise Exception(err)

    return decorator



import click

from .commands import to_csv, to_xlsx, to_json, compare
from .commons.consts import default_endpoints
from . import __version__


def print_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    click.echo(f"Version {__version__}")
    ctx.exit()


@click.group()
@click.option(
    "-v",
    "--version",
    is_flag=True,
    callback=print_version,
    expose_value=False,
    is_eager=True,
    help="Print out hycli version",
)
@click.option(
    "-e",
    "--endpoint-env",
    default="production",
    show_default=True,
    type=click.Choice(["localhost", "staging", "production"], case_sensitive=False),
    help="endpoint environment",
)
@click.option(
    "-u",
    "--username",
    envvar="HYCLI_USERNAME",
    default=None,
    help="your API username for accessing the environment",
)
@click.option(
    "-p",
    "--password",
    envvar="HYCLI_PASSWORD",
    default=None,
    help="your API password for accessing the environment",
)
@click.option(
    "--extractor", help="custom extractor endpoint",
)
@click.option(
    "-H",
    "--header",
    multiple=True,
    help="extractor endpoint header(s) can be multiple. Similair to curl, e.g. -H 'keyheader: value' -H '2ndkeyheader: 2ndvalue'",
)
@click.pass_context
def main(ctx, endpoint_env, username, password, extractor, header):
    """
    Can extract information from a directory of documents (invoices, receipts) to csv/xlsx using the Hypatos Extraction service.
    """
    ctx.obj = {
        "env": endpoint_env,
        "headers": {h.split(":")[0].strip(): h.split(":")[1].strip() for h in header},
    }

    ctx.obj["endpoints"] = {
        k: v for k, v in default_endpoints[endpoint_env].items() if v is not None
    }

    if username:
        ctx.obj["username"] = username

    if password:
        ctx.obj["password"] = password

    if extractor:
        ctx.obj["endpoints"]["extractor"] = extractor


main.add_command(to_csv.to_csv)
main.add_command(to_xlsx.to_xlsx)
main.add_command(to_json.to_json)
main.add_command(compare.compare)

if __name__ == "__main__":
    main()



import click
from halo import Halo

from .. import convert_to_xlsx, Services
from ..convert.commons import get_output_path


@click.command(context_settings=dict(max_content_width=200))
@click.argument("path", type=click.Path(exists=True))
@click.option("-w", "--workers", default=1, show_default=True, help="amount of workers")
@click.option(
    "-o",
    "--output",
    help="output path for xlsx file relative from current location (ends in .xlsx, overwrites previous)",
)
@click.option(
    "-n",
    "--normalize",
    is_flag=True,
    help="Normalize the probabilities to range(min, max) for every column",
)
@click.pass_context
def to_xlsx(ctx, path, workers, output, normalize: bool):
    """ Convert invoice to xlsx """
    # Services available
    spinner = Halo(spinner="dots")
    spinner.start()
    services = Services(ctx, check_up=True)
    spinner.succeed(f"Found dir: {path}\n ")

    # Get absolute output path
    output_path = get_output_path(path, output)

    # Conversion
    convert_to_xlsx(path, services, workers, output_path, normalize)

    spinner.succeed("Converted invoice(s) to xlsx")



import requests

import click

from halo import Halo

from .requests import request_token


class Services(object):
    def __init__(self, ctx, check_up=False):
        """
        Holds attributes related to service.
        """
        endpoints = ctx.obj.get("endpoints")

        self.env = ctx.obj.get("env")
        self.headers = ctx.obj.get("headers")
        self.username = ctx.obj.get("username")
        self.password = ctx.obj.get("password")
        self.authentication_endpoint = endpoints.get("authentication")

        if check_up:
            self.are_up(endpoints, ctx)

        self.extractor_endpoint = endpoints.get("extractor")

        self.token = None
        self.token_counter = 0
        self.refresh_token()

    def are_up(self, endpoints, ctx):
        """ Check if endpoints are reachable. """
        for name, url in endpoints.items():
            url_spinner = Halo(spinner="dots")
            svc = name.capitalize()

            try:
                req = requests.get(url)

            # On connection error.
            except requests.exceptions.ConnectionError:
                url_spinner.fail(f"{svc} on endpoint {url} is offline\n")
                endpoints[name] = None
                if name == "extractor":
                    click.echo(
                        f"Expected extraction service on {url}, service is mandatory."
                    )
                    ctx.exit()
            except Exception as e:
                url_spinner.fail(f"{svc} on endpoint {url} is offline\n")
                endpoints[name] = None
                if name == "extractor":
                    click.echo(
                        f"Expected extraction service on {url}, service is mandatory.\n{e}"
                    )
                    ctx.exit()
            # Connection was successful on 401 and 405.
            else:
                if req.status_code == 401 or req.status_code == 405:
                    url_spinner.succeed(f"{svc} endpoint is online\n")
                else:
                    url_spinner.fail(f"{svc} endpoint returns {req.status_code}\n")
                    endpoints[name] = None

    def get_token(self, refresh_on_every=50):
        if self.token_counter % refresh_on_every == 0:
            self.refresh_token()
        self.token_counter += 1
        return self.token

    def refresh_token(self):
        if self.env != "localhost":
            self.token = request_token(
                self.authentication_endpoint, self.username, self.password
            )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hycli', 'hycli.commands']

package_data = \
{'': ['*'], 'hycli': ['commons/*', 'convert/*', 'services/*']}

install_requires = \
['Jinja2>=2.11.2,<3.0.0',
 'XlsxWriter>=1.2.7,<2.0.0',
 'click-log>=0.3.2,<0.4.0',
 'click>=7.0,<8.0',
 'filetype>=1.0.5,<2.0.0',
 'halo>=0.0.29,<0.0.32',
 'openpyxl>=3.0.5,<4.0.0',
 'pandas>=1.0.5,<2.0.0',
 'requests>=2.22.0,<3.0.0',
 'xlrd>=1.2.0,<2.0.0']

entry_points = \
{'console_scripts': ['hycli = hycli.cli:main']}

setup_kwargs = {
    'name': 'hycli',
    'version': '0.3.28',
    'description': 'Interface package to convert invoice to xml by requesting different Hypatos services.',
    'long_description': None,
    'author': 'Dylan Bartels',
    'author_email': 'dylan.bartels@hypatos.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.12,<4.0.0',
}


setup(**setup_kwargs)
